<?php

namespace Drupal\danse\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\danse\RecipientSelectionManager;
use Drupal\danse\Service;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure Subscription Audit Notification Event settings for this site.
 */
final class Settings extends ConfigFormBase {

  /**
   * @var \Drupal\danse\Service
   */
  protected $danse;

  /**
   * @var \Drupal\danse\RecipientSelectionManager
   */
  protected $recipientSelectionManager;

  /**
   * Settings constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   * @param \Drupal\danse\Service $danse
   * @param \Drupal\danse\RecipientSelectionManager $recipient_selection_manager
   */
  public function __construct(ConfigFactoryInterface $config_factory, Service $danse, RecipientSelectionManager $recipient_selection_manager) {
    parent::__construct($config_factory);
    $this->danse = $danse;
    $this->recipientSelectionManager = $recipient_selection_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('danse.service'),
      $container->get('danse.recipient.selection.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'danse_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames(): array {
    return ['danse.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {

    $form['subscriptions_as_tab'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Subscriptions as tab in user profile'),
      '#default_value' => $this->config('danse.settings')->get('subscriptions_as_tab'),
      '#description' => $this->t('By default, subscriptions will be available in the user profile form. If you select this checkbox, the subscriptions will be made available in their own tab in the user profile instead.'),
    ];
    $form['recipient_selection_plugin'] = [
      '#type' => 'details',
      '#title' => $this->t('Recipient selection plugins'),
      '#open' => TRUE,
      '#tree' => TRUE,
      '#description' => $this->t('Please select all the recipient selection plugins for each of the event types. Subscribers will be notified as well and no user will be notified more than once, regardless how many plugins you select.'),
    ];
    foreach ($this->danse->getPluginInstances() as $id => $plugin) {
      $form['recipient_selection_plugin'][$id] = [
        '#type' => 'select',
        '#title' => $plugin->label(),
        '#options' => $this->recipientSelectionManager->pluginListForSelect(),
        '#default_value' => explode(',', $this->config('danse.settings')
          ->get('recipient_selection_plugin.' . $id)),
        '#required' => FALSE,
        '#multiple' => TRUE,
      ];
    }
    $form['prune'] = [
      '#type' => 'details',
      '#title' => $this->t('Settings to prune database records'),
      '#open' => FALSE,
      '#tree' => TRUE,
      '#description' => $this->t('Select, how many or for how long the records of each event type should be kept in the database.'),
    ];
    foreach ($this->danse->getPluginInstances() as $id => $plugin) {
      $form['prune'][$id] = [
        '#type' => 'fieldset',
        '#title' => $plugin->label(),
        '#attributes' => [
          'class' => [
            'container-inline',
            'fieldgroup',
            'form-composite',
          ],
        ],
      ];
      $form['prune'][$id]['value'] = [
        '#type' => 'number',
        '#min' => 0,
        '#size' => 5,
        '#default_value' => $this->config('danse.settings')->get('prune.' . $id . '.value'),
        '#required' => FALSE,
        '#states' => [
          'invisible' => [
            'select[name="prune[' . $id . '][type]"]' => ['value' => 'all'],
          ],
        ],
      ];
      $form['prune'][$id]['type'] = [
        '#type' => 'select',
        '#options' => [
          'all' => $this->t('All / forever'),
          'records' => $this->t('Database records'),
          'days' => $this->t('Days'),
          'weeks' => $this->t('Weeks'),
          'months' => $this->t('Months'),
          'years' => $this->t('Years'),
        ],
        '#default_value' => $this->config('danse.settings')->get('prune.' . $id . '.type'),
        '#required' => FALSE,
      ];

    }
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    /** @var \Drupal\Core\Config\Config $config */
    if ($config = $this->config('danse.settings')) {
      $config->set('subscriptions_as_tab', $form_state->getValue('subscriptions_as_tab'));
      foreach ($this->danse->getPluginInstances() as $id => $plugin) {
        $config->set('recipient_selection_plugin.' . $id, implode(',', $form_state->getValue(['recipient_selection_plugin', $id])));
        $config->set('prune.' . $id . '.value', $form_state->getValue(['prune', $id, 'value']));
        $config->set('prune.' . $id . '.type', $form_state->getValue(['prune', $id, 'type']));
      }
      $config->save();
    }
    parent::submitForm($form, $form_state);
  }

}
